/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann@t-online.de
 Latest version generated:	30.04.2007
 Filename:					RealTimeClock.c
 Description:    			Real Time Clock Implementation

 Note: 
 Real Time Clock Implementation with IC RTC4513 by EPSON
 
 Public Functions:
	RTC_Init
	RTC_SetTimeDate
	RTC_GetTimeDate
	RTC_SetDayOfWeek
	RTC_GetDayOfWeek

 Private Functions:
	RTC_WriteNibble
	RTC_ReadNibble
	RTC_WriteByte
	RTC_WriteBytes2Address
	RTC_ReadBytes2Address
	RTC_Set_Registers
	RTC_Byte2BCD
	RTC_BCD2Byte

 *******************************************************/
#include <avr/io.h>

#include "RealTimeClock.h"
#include "Main.h"
/*
** Definition of Ports and Data
*/
#define	RTC_Read	PINA
#define	RTC_Write	PORTA
#define	RTC_DDR		DDRA
/*
** Note on Input/Output of RTC_Data:
** Output (RTC_Data) of RTC4513 is Tristate,
** when RTC_Enable = Low level
*/
#define RTC_Data	PINA0  // Bidirectional pin
#define RTC_Clock	PINA1  // Output pin
#define RTC_Enable	PINA2  // Output pin
/*
** Definition of registers of RTC IC:
** SEC1REG	= Start of Sec Min Hour register
** DAY1REG	= Start of Day Month Year egister
** DOWREG	= Start od Day of Week register
** CDREG	= CD register
** CFREG	= CF register
*/
#define SEC1REG			0
#define DAY1REG			6
#define DOWREG			12
#define CDREG			13
#define CFREG			15

#define RTC_RESET		0	// used in RTC_Set_Registers
#define RTC_SET_MODE	1
/*
** Private Functions Prototypes
*/
void 	RTC_WriteNibble (uint8_t data);
uint8_t RTC_ReadNibble (void);
void 	RTC_WriteByte (uint8_t data);
void 	RTC_WriteBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count);
void 	RTC_ReadBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count);
void 	RTC_Set_Registers (uint8_t mode);
uint8_t	RTC_Byte2BCD (uint8_t data);
uint8_t	RTC_BCD2Byte (uint8_t bcd);

/*******************************************************
 Function: RTC_Init
	
 Purpose: Initialise the Real Time Clock
 Input parameters: None
 Return value: None

 *******************************************************/
void RTC_Init (void)
	{
/*
** Set Clock and Enable Pins to Output
*/
	SET_BIT(RTC_DDR, RTC_Clock);
	SET_BIT(RTC_DDR, RTC_Enable);
/*
** Set Data Pin to Input
*/
	CLEAR_BIT(RTC_DDR, RTC_Data);
/*
** Set Clock line to Low Level
** Set Data line to Low Level = Tristate of RTC4513
*/
	CLEAR_BIT(RTC_Write, RTC_Clock);
	CLEAR_BIT(RTC_Write, RTC_Data);
/*
** Initiate the RTC IC registers (CDREG and CFREG)
** Set 24 hour mode, set clock range from sec to year
*/
    RTC_Set_Registers (RTC_SET_MODE);
	}
/*******************************************************
 Function: RTC_Set_Registers
	
 Purpose: Initiate the RTC registers

 Input parameters:
 	uint8_t		RTC_RESET	: Reset clock, Stop clock, 24 hour mode
	uint8_t		RTC_SET_MODE: Set 24 hour mode, set clock
								range from sec to year
 
 Return value: None

 *******************************************************/
void RTC_Set_Registers (uint8_t mode)
	{
/*
** Set CF and CD registers of RTC IC
*/
	if (mode == RTC_SET_MODE)
		{
/*
** Set CF register bit D2 to 24 hour mode
*/
		SET_BIT(RTC_Write, RTC_Enable);
		RTC_WriteNibble (0x03);
		RTC_WriteNibble (CFREG);
		RTC_WriteNibble (0x04);		// Set to 24 hour mode
		CLEAR_BIT(RTC_Write, RTC_Enable);
/*
** Set CAL/HW bit in CD register
** Clock runs from seconds to year
*/
		SET_BIT(RTC_Write, RTC_Enable);
		RTC_WriteNibble (0x03);
		RTC_WriteNibble (CDREG);
		RTC_WriteNibble (0x02);  // Set CAL/HW (clock range switching bit)
		CLEAR_BIT(RTC_Write, RTC_Enable);
		}

	if (mode == RTC_RESET)
		{
/*
** Set CF register of RTC IC
** 	D0 = Reset clock
** 	D1 = Stop clock
** 	D2 = 24 hour mode
*/
		SET_BIT(RTC_Write, RTC_Enable);
		RTC_WriteNibble (0x03);
		RTC_WriteNibble (CFREG);
		RTC_WriteNibble (0x07);
		CLEAR_BIT(RTC_Write, RTC_Enable);
		}
	}
/*******************************************************
 Function: RTC_WriteNibble
	
 Purpose: Write a nibble (4 bit) to the RTC
 			4 bit in low Nibble of parameter data

 Input parameters:
 	uint8_t		data	(nibble in low 4 bits)
 
 Return value: None

 *******************************************************/
void RTC_WriteNibble (uint8_t data)
	{
	uint8_t		i;
/*
** First set data line to output
*/
	SET_BIT(RTC_DDR, RTC_Data);
/*
** Write the 4 data bits
*/
	for(i=0;i<4;i++)
		{
		if (((data>>i)&0x01) == 1)
			{
			SET_BIT(RTC_Write, RTC_Data);
			}
		SET_BIT(RTC_Write, RTC_Clock);
		CLEAR_BIT(RTC_Write, RTC_Clock);
		CLEAR_BIT(RTC_Write, RTC_Data);
    	}
/*
** Set data line to input
*/
	CLEAR_BIT(RTC_DDR, RTC_Data);
	}
/*******************************************************
 Function: RTC_ReadNibble
	
 Purpose: Read a nibble (4 bit) from the RTC
 			4 bit in high Nibble of para data

 Input parameters:
 	uint8_t		data	(nibble in high 4 bits)
 
 Return value: None

 *******************************************************/
uint8_t RTC_ReadNibble (void)
	{
	uint8_t		i;
	uint8_t		Byte=0;

	for(i=0;i<4;i++)
		{
		SET_BIT(RTC_Write, RTC_Clock);
		if (bit_is_set(RTC_Read, RTC_Data)>0)
			{
			SET_BIT(Byte, i);
			}

		CLEAR_BIT(RTC_Write, RTC_Clock);
    	}
    return Byte;
	}
/*******************************************************
 Function: RTC_WriteByte
	
 Purpose: Write a byte to the RTC

 Input parameters:
 	uint8_t		data	(the byte)
 
 Return value: None

 *******************************************************/
void RTC_WriteByte (uint8_t data)
	{
	RTC_WriteNibble (data);
	RTC_WriteNibble (data>>4);
	}
/*******************************************************
 Function: RTC_WriteBytes2Address
	
 Purpose: Write Count Bytes to an RTC Address

 Input parameters:
 	uint8_t		Addr	(where to start writing to RTC)
 	uint8_t	   *Data	(the Count bytes address)
 	uint8_t	    Count 	( Number of bytes to write) 

 Return value: None

 *******************************************************/
void RTC_WriteBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count)
	{
	uint8_t		i;

	SET_BIT(RTC_Write, RTC_Enable);
	RTC_WriteNibble (0x03);
	RTC_WriteNibble (Addr);

    for (i=0;i<Count;i++)
		{
		RTC_WriteByte (Data[i]);
		}
	CLEAR_BIT(RTC_Write, RTC_Enable);
	}
/*******************************************************
 Function: RTC_ReadBytes2Address
	
 Purpose: Read Count Bytes from an RTC Address

 Input parameters:
 	uint8_t		Addr	(where to start writing to RTC)
 	uint8_t	   *Data	(the Count bytes address)
 	uint8_t	    Count 	( Number of bytes to write) 

 Return value: None

 *******************************************************/
void RTC_ReadBytes2Address (uint8_t Addr, uint8_t *Data, uint8_t Count)
	{
	uint8_t		i;

	SET_BIT(RTC_Write, RTC_Enable);
	RTC_WriteNibble (0x0c);
	RTC_WriteNibble (Addr);

	for (i=0;i<Count;i++)
		{
        Data[i] = RTC_ReadNibble ();
        Data[i] = Data [i] | (RTC_ReadNibble () <<4);
		}
	CLEAR_BIT(RTC_Write, RTC_Enable);
	}
/*******************************************************
 Function: RTC_Byte2BCD
	
 Purpose: Convert Byte to BCD

 Input parameters:
 	uint8_t		data	(the byte to be converted)

 Return value: The converted byte

 *******************************************************/
uint8_t	RTC_Byte2BCD (uint8_t data)
	{
	return ((data/10)<<4) + (data%10);
	}
/*******************************************************
 Function: RTC_BCD2Byte
	
 Purpose: Convert BCD to Byte

 Input parameters:
 	uint8_t		data	(the BCD to be converted)

 Return value: The converted BCD

 *******************************************************/
uint8_t	RTC_BCD2Byte (uint8_t bcd)
	{
	return (bcd>>4)*10 + (bcd&0x0f);
	}
/*******************************************************
 Function: RTC_SetTimeDate 
	
 Purpose: Set the RTC Time or Date

 Input parameters:

 	case: para = RTC_SET_TIME
 	uint8_t		p1 = Second {0-59}
 	uint8_t		p2 = Minute {0-59}
 	uint8_t		p3 = Hour 	{0-23}

 	case: para = RTC_SET_DATE
 	uint8_t		p1 = Day	{1-31}
 	uint8_t		p2 = Month	{1-12}
 	uint8_t		p3 = Year	{0-xx} (without 2000)

 Return value: None

 *******************************************************/
void RTC_SetTimeDate (uint8_t p1, uint8_t p2, uint8_t p3, uint8_t para)
	{
	uint8_t 	data[3];
	uint8_t 	reg = SEC1REG;

	if (para == RTC_SET_DATE) reg = DAY1REG;

    data[0] = RTC_Byte2BCD (p1); // second or day
    data[1] = RTC_Byte2BCD (p2); // minute or month
    data[2] = RTC_Byte2BCD (p3); // hour   or year
	RTC_Set_Registers (RTC_RESET);
    RTC_WriteBytes2Address (reg, data, 3);
	RTC_Set_Registers (RTC_SET_MODE);
	}
/*******************************************************
 Function: RTC_SetDayOfWeek
	
 Purpose: Set the Day of Week of the RTC

 Input parameters:
 	uint8_t		DOW		(Day of week: 0=Monday, 6=Sunday)

 Return value: None

 *******************************************************/
void RTC_SetDayOfWeek (uint8_t DOW)
	{
	RTC_Set_Registers (RTC_RESET);	
	SET_BIT(RTC_Write, RTC_Enable);
	RTC_WriteNibble (0x03);
	RTC_WriteNibble (DOWREG);
	RTC_WriteNibble (DOW);
	CLEAR_BIT(RTC_Write, RTC_Enable);
	RTC_Set_Registers (RTC_SET_MODE);
	}
/*******************************************************
 Function: RTC_GetDayOfWeek
	
 Purpose: Read the day of the week

 Input parameters: None

 Return value: uint8_t Day of week

 *******************************************************/
uint8_t RTC_GetDayOfWeek (void)
	{
	uint8_t		i;
/*
** Read again if fr=1 of low nibble (carry read flag)
*/
	do
		{
		RTC_ReadBytes2Address (DOWREG, &i, 1);
		} while (i&0x08);

	return i;
	}
/*******************************************************
 Function: RTC_GetTimeDate
	
 Purpose: Get the RTC Time or Date

 Input parameters:
 	uint8_t		type:	either ('s'-sec, 'm'-min, 'h'-hrs)
						or     ('D'-day, 'M'-mon, 'Y'-yrs)
 	uint8_t		para:
				= RTC_READ_TIME		: Read the Time from the RTC
				= RTC_TAKE_READ_TIME: Take the earlier read Time

				= RTC_READ_DATE		: Read the Date from the RTC
				= RTC_TAKE_READ_DATE: Take the earlier read Date

 Return value: The time/date: either
 				sec, min, hrs	or
 				Day, Month, Year (without 2000)

 *******************************************************/
uint8_t	RTC_GetTimeDate (uint8_t type, uint8_t para)
	{
	static uint8_t	data[3];
	uint8_t 		i = SEC1REG;
/*
** Read again if fr=1 of low nibble (carry read flag)
*/
    if (para == RTC_READ_TIME || para == RTC_READ_DATE)
		{
		if (para == RTC_READ_DATE) i = DAY1REG;

        do
			{
       		RTC_ReadBytes2Address (i, data, 3);
       		} while (data[1]&0x80 || data[2]&0x80);
    	}
/*
** Extract required result
*/
	switch (type)
		{
		case 'D': 
			i = RTC_BCD2Byte (data[0]&0x3f);
			break;
		case 'M':
			i = RTC_BCD2Byte (data[1]&0x1f);
			break;
		case 'Y':
			i = RTC_BCD2Byte (data[2]);
			break;
		case 's':
			i = RTC_BCD2Byte (data[0]&0x7f);
			break;
		case 'm':
			i = RTC_BCD2Byte (data[1]&0x7f);
			break;
		case 'h': 
			i = RTC_BCD2Byte (data[2]&0x3f);
			break;
		}
	return i;
	}
